import aesara.tensor.rewriting.basic
import aesara.tensor.rewriting.elemwise
import aesara.tensor.rewriting.extra_ops

# Register JAX specializations
import aesara.tensor.rewriting.jax
import aesara.tensor.rewriting.math
import aesara.tensor.rewriting.shape
import aesara.tensor.rewriting.special
import aesara.tensor.rewriting.subtensor
import aesara.tensor.rewriting.uncanonicalize
