.. _css:

.. raw:: html

    <style type="text/css">
    .black {color:black};
    .blue  {color:blue};
    .red   {color:red};
    .green {color:green};
    .pink  {color:pink};
    </style>

.. role:: black
.. role:: blue
.. role:: red
.. role:: green
.. role:: pink
